from __future__ import annotations

from functools import cached_property
from typing import Any, List, Optional

import numpy as np
import xarray as xr

from mfire.settings import get_logger
from mfire.text.synthesis.builder import SynthesisBuilder
from mfire.text.synthesis.reducer import SynthesisReducer
from mfire.text.synthesis.wind_reducers.base_param_summary_builder import (
    BaseParamSummaryBuilder,
)
from mfire.text.synthesis.wind_reducers.wind import WindSummaryBuilder
from mfire.utils.jinja_template import JinjaTemplateCreator
from mfire.utils.template import Template

from .wind_reducers.gust_summary_builder import GustSummaryBuilder

# Logging
LOGGER = get_logger(name=__name__, bind="wind_synthesis")


class WindReducer(SynthesisReducer):
    """Reducer class for the wind force and the gust force."""

    @staticmethod
    def _process_dataset_valid_time(dataset: xr.Dataset) -> Optional[xr.Dataset]:
        """Process the valid_time of the input Dataset.

        The first term of valid_time is removed. So if there is less than 2 terms,
        then it remains no data and None is returned. Else, a Dataset with a reduced
        valid_time dim is returned.
        """
        valid_time: np.ndarray[np.datetime64] = dataset.valid_time.values

        if len(valid_time) <= 1:
            return

        dataset = dataset.isel(valid_time=range(1, len(valid_time)))

        dataset["previous_time"] = xr.DataArray(
            data=valid_time[:-1],
            coords=[dataset.valid_time.values],
            dims=["valid_time"],
        )

        return dataset

    def _compute(self) -> dict:
        """Compute the summary."""
        self.reduction = {}

        dataset: Optional[xr.Dataset] = self._process_dataset_valid_time(
            self.parent_data
        )

        # Get Gust and Wind summary
        for summary_builder_class in [WindSummaryBuilder, GustSummaryBuilder]:
            # If there is no data, then call no_data_summary method
            if dataset is None:
                self.reduction.update(summary_builder_class.no_data_summary())
                continue

            builder: BaseParamSummaryBuilder = summary_builder_class(
                self.parent, dataset
            )
            param_summary: dict = builder.compute(
                reference_datetime=self.parent.parent.production_datetime
            )
            self.reduction.update(param_summary)

        return self.reduction


class WindBuilder(SynthesisBuilder):
    """WindBuilder class."""

    reducer_class: type = WindReducer

    @property
    def template_name(self) -> str:
        """Get the template name."""
        return "wind"

    @property
    def params(self) -> tuple[str, str]:
        """Get the name of the parameters."""
        return "wind", "gust"

    @cached_property
    def cases(self) -> dict:
        """Get the reduction case of the parameters."""
        cases: dict[str, Optional[str]] = {}
        for param in self.params:
            try:
                cases[param] = self.reduction[param]["case"]
            except KeyError:
                cases[param] = None
                LOGGER.error(f"Case not found for '{param}' param !")
        return cases

    @cached_property
    def template_key(self) -> list[tuple[str, str]]:
        """Get the template key from the selector key."""
        return [(param, self.cases[param]) for param in self.params]

    @cached_property
    def template(self) -> Optional[str | List[str]]:
        """
        Retrieve the template from the file system.

        Returns:
            str: The template or None if the template name is not set or the template
                was not found.
        """
        if len(self.cases) != 2:
            return None

        for param, case in self.cases.items():
            if self.template_retriever.table[param].get(case) is None:
                LOGGER.error(f"WindBuilder do not found the template for '{case}' !")
                return None

        return [self.template_retriever.get(key) for key in self.template_key]

    def _template_format(self, template: str, formating_data) -> Template:
        """Format the template with data given as kwargs."""
        jinja_template = JinjaTemplateCreator().run(template)
        return Template(jinja_template.render(**formating_data))

    def _generate_text(self, template: list[str], reduction: Any) -> None:
        """Generate the text from a list of templates and some formating data.

        If the wind is 'faible' (case 1) and there is gust (case > 0), then
        remove the wind summary in reduction
        """
        params: tuple = self.params
        if self.cases["wind"] == "1" and self.cases["gust"] != "0":
            params = ("gust",)
            template.pop(self.params.index("wind"))
            reduction.pop("wind")
        reduction = [reduction[param] for param in params]

        # Generate the text from template and formating_data
        self.text = " ".join(
            self._template_format(text, text_reduction)
            for text, text_reduction in zip(template, reduction)
        )
